import csv
from src.setup.Settings import read_config, write_config
from src.datastore.Memo import memo, add_to_memo, format_pid, retrieve_from_memo, update_memo


# Read a supplied source file (csv or txt)
# and add all required records to the record memo
# with the status 'pending'.
def read_input_list(source_file):
    endpoint = read_config("endpoint")
    label = read_config("corefile")

    if source_file.endswith(".csv"):
        import_list_from_csv(source_file, endpoint, label)
    elif source_file.endswith(".txt"):
        import_list_from_txt(source_file, endpoint, label)


def import_list_from_csv(source_file, endpoint, label):
    with open(source_file, newline="", encoding="utf-8") as f:
        source_data = csv.DictReader(f, delimiter=",")

        for row in source_data:
            record_irn = int(row.get("record_irn"))
            media_irn = int(row.get("media_irn"))
            if read_config("use_skipfile"):
                if skip_check(record_irn):
                    break

            record_pid = add_to_memo(irn=record_irn, endpoint=endpoint, label=label)
            if media_irn:
                update_memo(record_pid, "media", media_irn)


def import_list_from_txt(source_file, endpoint, label):
    with open(source_file, "r", encoding="utf-8") as f:
        source_data = f.readlines()

        for row in source_data:
            record_irn = int(row.strip())
            if read_config("use_skipfile"):
                if skip_check(record_irn):
                    break

            add_to_memo(irn=record_irn, endpoint=endpoint, label=label)


def skip_check(irn):
    if irn in read_config("skiplist"):
        return True

    return False


def add_irn_to_memo(record_irn=None, media_irn=None, endpoint=None, label=None):
    # If the record is not already in the memo, add it now
    record_pid = format_pid(endpoint=endpoint, irn=record_irn)
    if not retrieve_from_memo(record_pid):
        add_to_memo(status="pending", irn=record_irn, endpoint=endpoint, label=label)

    # If specific images are required, append their IRNs to the memo
    if media_irn:
        update_memo(record_pid, "media", media_irn)



import yaml
from datetime import date
from src.setup.Settings import read_config, write_config
from src.monitoring.Stats import stats


def generate_export_report():
	stats.process_runtimes()
	stats.count_new_and_modified()
	output_dir = read_config("output_dir")
	export_filename = read_config("export_filename")
	export_report_filename = "{}.yaml".format(export_filename)
	export_report_path = "{d}/{e}/{f}".format(d=output_dir,
	                                          e=export_filename,
	                                          f=export_report_filename)
	write_config("export_report_path", export_report_path)
	export_report_dict = gather_report_data()
	with open(export_report_path, "w", encoding="utf-8") as export_report:
		yaml.dump(export_report_dict, export_report)

	stats.print_stats()


def gather_report_data():
	export_date = date.today()
	report_dict = {"export_id": read_config("export_id"),
	               "runtime": stats.run_time,
	               "api_calls": stats.api_call_count,
	               "harvest_time": stats.harvest_time,
	               "extension_time": stats.extension_time,
	               "extension_records": stats.extension_records_count,
	               "processing_time": stats.processing_time,
	               "export_filenames": stats.export_filenames,
	               "records_written": stats.file_write_counts,
	               "new_record_counts": stats.new_record_count,
	               "update_counts": stats.modified_record_count,
	               "year": export_date.year,
	               "month": export_date.month,
	               "day": export_date.day}
	return report_dict


def analyse_export():
	output_dir = read_config("output_dir")
	export_report_path = read_config("export_report_path")
	if export_report_path:
		with open(export_report_path, "r", encoding="utf-8") as f:
			export_report = yaml.safe_load(f)
			export_files = export_report["export_filenames"]
			for file in export_files:
				export_file_path = "{d}/{f}".format(d=output_dir, f=file)




class RelationshipOutput(Output):
    def __init__(self, label=None, fieldnames=None, rules=None):
        super().__init__(label, fieldnames, rules)

        self.object = rules.object
        self.subject = rules.subject
        self.operator = rules.operator

        self.sets = {}

    def create_sets(self):
        object_records = self.find_objects()
        for object_record in object_records:
            relationship_set = []
            object_comparator_value = self.prepare_value_for_comparison(object_record, self.object)
            relationship_set.extend([record.data.get("pid") for record in object_records if
                                     self.check_subjects_for_inclusion(record, object_comparator_value) == True])
            if len(relationship_set) > 0:
                object_pid = object_record.data.get("pid")
                memo[object_pid]["relationship_found"] = True
                self.sets.update({object_pid: relationship_set})

    def find_objects(self):
        return [i for i in memo if self.label in memo[i]["write_to"]]

    def prepare_value_for_comparison(self, record, rule):
        comparator_value = literal(record.data, rule.get("path"))
        if comparator_value:
            comparator_modified = self.modify_comparator_value(rule, comparator_value)

            if read_config("strip_relationship_values"):
                return comparator_modified.strip()

    def modify_comparator_value(self, comparator_rule, comparator_value):
        modifier = comparator_rule.get("modifier")

        match modifier:
            case "trim":
                modifier_value = comparator_rule.get("modifier_value")
                try:
                    int(modifier_value)
                except TypeError:
                    print("Modifier value must be integer")
                    return None
                if modifier_value > 0:
                    return comparator_value[modifier_value, -1]
                elif modifier_value < 0:
                    return comparator_value[0, modifier_value]

            case "capitalise":
                return comparator_value.capitalize()

            case "equals":
                modifier_value = comparator_rule.get("modifier_value")
                if comparator_value == modifier_value:
                    return True
                else:
                    return False

            case _:
                return comparator_value

    def check_subjects_for_inclusion(self, subject_record, object_comparator_value):
        subject_pid = subject_record.data.get("pid")
        if not memo[subject_pid].get("relationship_found"):
            subject_comparator_value = self.prepare_value_for_comparison(subject_record, self.subject)

            match self.operator:
                case "equals":
                    if subject_comparator_value == object_comparator_value:
                        memo[subject_pid]["relationship_found"] = True
                        return True

    def create_rows_from_sets(self):
        for object_pid in list(self.sets.keys()):
            object_record = records.find_record(pid=object_pid)
            for subject_pid in self.sets[object_pid]:
                subject_record = records.find_record(pid=subject_record)
                self.create_relationship_row(object_record, subject_record)

    def create_relationship_row(self, object_record, subject_record):
        relationship_row = OutputRow(**kwargs)
        self.rows.append(relationship_row.values)

'''
- resourceRelationship:
    output_type: relationship
    relationship:
        object:
            label: object
            path: identifier
            modifier: trim
            modifier_value: -2
        subject:
            label: subject
            path: identifier
            modifier: trim
            modifier_value: -2
        operator: equals
    fields:
      - resourceRelationshipID:
          - format_string:
              string: tepapa:collection/relationship/{}/{}/{}
              path: object.id, subject.id, ordinal
      - resourceID:
          - literal:
              path: object.pid
      - relationshipOfResourceID:
          - hardcoded:
              value: "[URL FOR THIS REL]"
      - relatedResourceID:
          - literal:
              path: subject.pid
      - relationshipofResource:
          - hardcoded:
              value: Same specimen.
      - relationshipAccordingTo:
          - literal:
              path: object.identification.i.identifiedBy.title
      - relationshipEstablishedDate:
          - literal:
              path: object.identification.i.dateIdentified
      - relationshipRemarks:
          - hardcoded:
              value: null

'''


import os, shutil
from datetime import datetime
from src.setup.Settings import read_config


def create_export_dir():
	output_dir = read_config("output_dir")
	export_filename = read_config("export_filename")
	export_path = "{d}/{e}".format(d=output_dir, e=export_filename)
	if not os.path.exists(export_path):
		os.mkdir(export_path)


def validate_exports():
	pass


def zip_exported_files(export_files):
	# Currently produces invalid zip files, so not using
	from zipfile import ZipFile
	output_dir = read_config("output_dir")
	export_filename = read_config("export_filename")
	zipped_filename = "{}.zip".format(export_filename)
	write_location = "{d}/{e}/{f}".format(d=output_dir,
	                                      e=export_filename,
	                                      f=zipped_filename)
	with ZipFile(write_location, "w") as export_zip:
		for file in export_files:
			file_path = "{d}/{e}/{f}".format(d=output_dir,
	                                        e=export_filename,
	                                        f=file)
			export_zip.write(file_path)
	return write_location


def delete_old_exports():
	output_dir = read_config("output_dir")
	exports = os.listdir(output_dir)
	for item in exports:
		item_path = "{d}/{i}".format(d=output_dir,
		                             i=item)
		last_modified = os.path.getmtime(item_path)
		last_modified_dt = datetime.fromtimestamp(last_modified)
		if compare_timestamps(last_modified_dt):
			if os.path.isdir(item_path):
				shutil.rmtree(item_path)
			elif os.path.isfile(item_path):
				os.remove(item_path)


def compare_timestamps(last_modified):
	now = datetime.now()
	delta = now - last_modified
	if delta.days > 14:
		return True
	return False



from math import ceil
import time
from tqdm import tqdm
from askCO import Search, Scroll, Resource

from src.setup.Settings import read_config, write_config
from src.monitoring.Stats import stats
from src.datastore.RecordStore import records
from src.datastore.Record import Record
from src.datastore.Memo import memo, format_pid, retrieve_from_memo, add_to_memo, update_memo
from src.util.ProcessingFunctions import collate_list, literal


# Harvest records from Te Papa's Collections Online API
class Harvester():
    def __init__(self, sleep=0.1):
        self.quiet = read_config("quiet")
        self.sleep = sleep
        self.api_key = read_config("api_key")
        self.timeout = read_config("timeout")
        self.attempts = read_config("attempts")
        self.label = read_config("corefile")

        if not self.quiet:
            print("Harvester details: API key = {ak}, timeout = {t}, attempts = {a}".format(ak=self.api_key,
                                                                                            t=self.timeout,
                                                                                            a=self.attempts))

    def complete_harvest(self, mode):
        # Gather search or scroll parameters and turn returned results into Record objects
        endpoint = read_config("endpoint")
        query = read_config("query")
        sort = read_config("sort")
        filters = read_config("filters")
        start = 0
        size = read_config("size")

        stats.mark_time("harvest start")

        if mode == "search":
            self.harvest_search(endpoint, query, sort, filters, start, size)
        elif mode == "scroll":
            self.harvest_scroll(endpoint, query, sort, filters, size)

        stats.mark_time("harvest end")

        self.retrieve_extension_records()

    def retrieve_extension_records(self):
        extension_list = [memo[i] for i in memo.keys() if memo[i].get("is_extension")]

        stats.mark_time("extension start")

        if len(extension_list) > 0:
            self.run_extension_harvest(extension_list)

        stats.mark_time("extension end")

        print("All records saved")

    def harvest_search(self, endpoint, query, sort, filters, start, size):
        initial_search = Search(
            quiet=self.quiet,
            sleep=self.sleep,
            api_key=self.api_key,
            timeout=self.timeout,
            attempts=self.attempts,
            endpoint=endpoint,
            query=query,
            sort=sort,
            filters=filters,
            start=start,
            size=1)

        # Get result count to work out how many pages to request
        initial_search.send_query()
        stats.api_call_count += 1

        stats.initial_record_count = initial_search.record_count
        if not read_config("quiet"):
            print("Search record count: {}".format(stats.initial_record_count))

        if read_config("record_limit"):
            record_limit = read_config("record_limit")
        else:
            record_limit = stats.initial_result_count

        page_count = ceil(record_limit / size)

        # Request each page of results
        for i in tqdm(range(0, page_count), desc="Retrieving search pages: "):
            search_page = Search(
                quiet=self.quiet,
                sleep=self.sleep,
                api_key=self.api_key,
                timeout=self.timeout,
                attempts=self.attempts,
                endpoint=endpoint,
                query=query,
                sort=sort,
                filters=filters,
                start=start,
                size=size)

            search_page.send_query()
            stats.api_call_count += 1

            if len(search_page.records) > 0:
                for record in search_page.records:
                    if self.duplicate_check(record, endpoint):
                        self.save_record(record=record, endpoint=endpoint, label=self.label)

            start += size
            time.sleep(self.sleep)

    def harvest_scroll(self, endpoint, query, sort, filters, size):
        scroll = Scroll(
            quiet=self.quiet,
            sleep=self.sleep,
            api_key=self.api_key,
            timeout=self.timeout,
            attempts=self.attempts,
            endpoint=endpoint,
            query=query,
            sort=sort,
            size=size,
            filters=filters,
            duration=1,
            max_records=read_config("record_limit"))

        scroll.send_query()
        print("Scroll complete")

        stats.initial_record_count = scroll.record_count
        if not read_config("quiet"):
            print("Search record count: {}".format(stats.initial_record_count))

        stats.api_call_count += ceil(len(scroll.records) / size) + 1

        if scroll.record_count > 0:
            print("Saving scrolled records")
            for record in tqdm(scroll.records, desc="Saving scrolled records: "):
                if self.duplicate_check(record, endpoint):
                    self.save_record(record=record, endpoint=endpoint, label=self.label)

    def harvest_from_list(self):
        stats.mark_time("harvest start")
        stats.source_list_count = len(memo)
        endpoint = read_config("endpoint")

        print("Retrieving records from list")

        irn_batch = []

        # Segment list into batches of 250 and run a search for each batch
        for pid in list(memo.keys()):
            memo_record = retrieve_from_memo(pid)
            formatted_irn = "(id:{})".format(memo_record["irn"])
            if formatted_irn not in irn_batch:
                irn_batch.append(formatted_irn)

            if len(irn_batch) == 250:
                self.retrieve_and_save_batch(irn_batch, endpoint)
                irn_batch = []

        # Retrieve and save any remaining records
        if len(irn_batch) > 0:
            self.retrieve_and_save_batch(irn_batch, endpoint)

        stats.mark_time("harvest end")

        self.retrieve_extension_records()

    def retrieve_and_save_batch(self, irn_batch, endpoint):
        batch_records = self.batch_search(irn_batch, endpoint)
        if len(batch_records.records) > 0:
            for record in batch_records.records:
                self.save_record(record, endpoint, read_config("corefile"))

    def batch_search(self, irn_batch, endpoint):
        query = " OR ".join(irn_batch)

        search_batch = Search(
            quiet=self.quiet,
            sleep=self.sleep,
            api_key=self.api_key,
            timeout=self.timeout,
            attempts=self.attempts,
            endpoint=endpoint,
            query=query,
            size=250)
        search_batch.send_query()
        stats.api_call_count += 1

        return search_batch

    def duplicate_check(self, record, endpoint):
        # Check if record has already been received and update memo as needed
        memo_check = retrieve_from_memo(record["pid"])
        if memo_check:
            if memo_check["status"] == "received":
                return False
            elif memo_check["status"] == "pending":
                update_memo(record["pid"], "status", "received")
        else:
            add_to_memo(status="received", irn=record["id"], endpoint=endpoint, label=self.label)

        return True

    def save_record(self, record, endpoint, label):
        new_record = Record(data=record, endpoint=endpoint)
        records.append(new_record)
        if read_config("use_quality_score"):
            save_quality_score(record)
        if label == read_config("corefile"):
            self.check_for_triggers(new_record)

    def check_for_triggers(self, record):
        # Review a harvested record against mapping rules to see what extension records are needed
        mapping = read_config("mapping")
        for trigger in mapping.reharvest_triggers:
            if trigger.parent_endpoint == record.endpoint:
                self.run_extension_trigger(record=record, trigger=trigger)

    def run_extension_trigger(self, record, trigger):
        # Find all endpoints/irns for extension records and add to memo
        # Applies a label if the new record should be written out
        # Label is None if the new record is just for a lookup function
        paths = trigger.harvest_path.split(", ")
        for path in paths:
            trigger_path = path.split(".")
            if "i" in trigger_path:
                extension_irns = collate_list(record.data, path=trigger_path)
                if extension_irns:
                    extension_irns = [i for i in extension_irns if i is not None]
                    for extension_irn in extension_irns:
                        self.save_extension_details(irn=extension_irn, record=record, trigger=trigger)

            else:
                extension_irn = literal(record.data, path=trigger_path)
                if extension_irn:
                    self.save_extension_details(irn=extension_irn, record=record, trigger=trigger)

    def save_extension_details(self, irn, record, trigger):
        extension_pid = format_pid(trigger.harvest_endpoint, irn)

        # Check if record is already in memo
        if retrieve_from_memo(extension_pid):
            # If so, add new label if record will be written out
            if trigger.label:
                update_memo(extension_pid, "add_label", trigger.label)
                update_memo(extension_pid, "extends", record.pid, label=trigger.label)
        else:
            # Otherwise add to memo
            add_to_memo(status="pending",
                        irn=irn,
                        endpoint=trigger.harvest_endpoint,
                        extension=True,
                        label=trigger.label)
            update_memo(extension_pid, "extends", record.pid, label=trigger.label)

    def run_extension_harvest(self, extension_list):
        # Work through list of triggered queries to pull down extra records
        print("Running extra queries")

        # Split up records by endpoint to allow searching just by IRN
        queries_by_endpoint = {"agent": [],
                               "fieldcollection": [],
                               "object": [],
                               "place": [],
                               "taxon": []}

        for pending_record in tqdm(extension_list, desc="Working: "):
            endpoint = pending_record["endpoint"]
            irn = pending_record["irn"]
            pid = format_pid(endpoint, irn)
            if retrieve_from_memo(pid).get("status") == "received":
                continue
            else:
                # Format a segment of the search query and store it to run a batch search
                irn_formatted = "(id:{})".format(irn)

                if irn_formatted not in queries_by_endpoint[endpoint]:
                    queries_by_endpoint[endpoint].append(irn_formatted)

                # Run whenever there's 250 IRNs saved for an endpoint
                if len(queries_by_endpoint[endpoint]) == 250:
                    extension_batch = self.batch_search(irn_batch=queries_by_endpoint[endpoint], endpoint=endpoint)
                    self.save_extension_records(extension_results=extension_batch.records, extension_endpoint=endpoint)
                    queries_by_endpoint[endpoint] = []

        # Run searches for remaining records
        for endpoint in list(queries_by_endpoint.keys()):
            if len(queries_by_endpoint[endpoint]) > 0:
                extension_batch = self.batch_search(irn_batch=queries_by_endpoint[endpoint], endpoint=endpoint)
                self.save_extension_records(extension_results=extension_batch.records, extension_endpoint=endpoint)

    def save_extension_records(self, extension_results, extension_endpoint):
        for record in extension_results:
            pid = record.get("pid")

            if retrieve_from_memo(pid).get("status") == "pending":
                if self.duplicate_check(record, extension_endpoint):
                    self.save_record(record=record, endpoint=extension_endpoint, label=None)
                    stats.extension_records_count += 1


def save_quality_score(record):
    try:
        record_quality_score = record["_meta"]["qualityScore"]
        if not stats.quality_score_lower:
            stats.quality_score_lower = record_quality_score
        if not stats.quality_score_upper:
            stats.quality_score_upper = record_quality_score
        if record_quality_score < stats.quality_score_lower:
            stats.quality_score_lower = record_quality_score
        if record_quality_score > stats.quality_score_upper:
            stats.quality_score_upper = record_quality_score
    except KeyError:
        pass



from src.setup.Settings import read_config, write_config
from src.monitoring.Stats import stats
from src.datastore.RecordStore import records
from src.datastore.Memo import memo
from src.util.Processor import RowProcessor
from src.util.ProcessingFunctions import collate_list, literal, step_to_field
from src.util.Writer import WriteFile
from tqdm import tqdm


class Output():
    def __init__(self, label=None, endpoint=None, fieldnames=None, reference_column=None,
                 rules=None, explode_on=None, reduce_on=None, requires=None):
        self.label = label
        self.endpoint = endpoint
        self.fieldnames = fieldnames
        self.reference_column = reference_column
        self.rules = rules
        self.explode_on = explode_on
        self.reduce_on = reduce_on
        self.requires = requires
        self.check_requirements = False

        stats.file_write_counts.update({label: 0})

        self.rows = []

        self.output_file = None

        if not read_config("quiet"):
            print("Prepping output for {} file".format(self.label))

    def write_output(self):
        self.output_file = self.create_output_file()
        self.prepare_header_row()
        self.prepare_output()
        self.output_file.write_rows(self.rows, self.fieldnames)

    def create_output_file(self):
        print("Writing {} file".format(self.label))
        return WriteFile(self.label)

    def prepare_header_row(self):
        header_row = [i for i in self.fieldnames]
        if self.reference_column:
            header_row.insert(0, self.reference_column)
        self.output_file.write_header_row(header_row)

    def prepare_output(self):
        if read_config("restrict_locality") and self.endpoint == "object":
            remove_restricted_localities()

        self.process_records_to_rows()

    def process_records_to_rows(self):
        # Processes each included record to get printable values
        # Appends to self.rows list before writing all records
        output_record_pids = [i for i in memo if self.label in memo[i]["write_to"]]

        if not read_config("quiet"):
            print("Writing {} records to file".format(len(list(output_record_pids))))

        for record_pid in output_record_pids:
            record = records.records.get(record_pid)
            if record:
                # Check if the record's rows need to point to another record
                pointers = memo[record_pid]["structure"][self.label]["extends"]
                if len(pointers) > 0:
                    if not read_config("quiet"):
                        print("{r} has pointers: {p}".format(r=record_pid, p=pointers))

                    for pointer in pointers:
                        self.chop_up_record(record=record, pointer=pointer)
                else:
                    self.chop_up_record(record=record)
            else:
                if not read_config("quiet"):
                    print("No record with pid {} found in saved records".format(record_pid))

        print("Processing values")

        for output_row in tqdm(self.rows, desc="Working: "):
            RowProcessor(self.rules, output_row, self.requires)

    def chop_up_record(self, record=None, pointer=None):
        kwargs = {"data": None,
                  "pointer": pointer,
                  "rules": self.rules,
                  "record_pid": record.pid,
                  "explode_on": self.explode_on,
                  "explode_ordinal": None,
                  "requires": self.requires,
                  "group_role": None}

        if self.reference_column:
            kwargs["write_pointer"] = True

        if read_config("mode") == "list":
            record = remove_nonwriting_images(record)

        # Cuts down the record to a subset if needed
        if self.reduce_on:
            reduce_path = self.reduce_on.split(".")
            data = step_to_field(record.data, reduce_path)
        else:
            data = record.data

        # Checks if the record explodes at the same level it's reduced to
        explode_reduce = False
        if self.reduce_on == self.explode_on:
            explode_reduce = True

        if data:
            # Runs if the record should be checked for multiple rows
            if self.explode_on:
                # Works out how many rows should be written
                if not explode_reduce:
                    explode_data = step_to_field(data, self.explode_on)
                    if explode_data:
                        explode_size = len(explode_data)
                    else:
                        return None
                else:
                    explode_size = len(data)

                # Runs if there are multiple rows to write out
                if explode_size > 1:
                    # Runs if multiple rows should be grouped with a parent row
                    if read_config("group_rows"):
                        if explode_reduce:
                            kwargs["data"] = data[0]
                        else:
                            kwargs["data"] = data
                        kwargs["group_role"] = "parent"
                        kwargs["explode_ordinal"] = 0
                        # Writes parent row
                        self.rows.append(OutputRow(**kwargs))
                        for i in range(0, explode_size):
                            if explode_reduce:
                                kwargs["data"] = data[i]
                            else:
                                kwargs["data"] = data
                            kwargs["group_role"] = "child"
                            kwargs["explode_ordinal"] = i
                            # Writes each child row
                            self.rows.append(OutputRow(**kwargs))
                    # Runs if multiple rows don't need to be grouped
                    else:
                        for i in range(0, explode_size):
                            if explode_reduce:
                                kwargs["data"] = data[i]
                            else:
                                kwargs["data"] = data
                            kwargs["group_role"] = None
                            kwargs["explode_ordinal"] = i
                            self.rows.append(OutputRow(**kwargs))
                # Runs if there is only one row needed
                else:
                    if explode_reduce:
                        kwargs["data"] = data[0]
                    else:
                        kwargs["data"] = data
                    kwargs["group_role"] = None
                    kwargs["explode_ordinal"] = 0
                    self.rows.append(OutputRow(**kwargs))

            # Runs if the record isn't being exploded
            else:
                kwargs["data"] = data
                kwargs["group_role"] = None
                kwargs["explode_ordinal"] = None
                self.rows.append(OutputRow(**kwargs))

        else:
            if not read_config("quiet"):
                print("Record {p} has no data for {l} output".format(p=record.pid, l=self.label))

    def check_if_row_required(self, record_data):
        for requirement in self.requires:
            req_path = requirement["path"]
            req_value = requirement["value"]
            record_value = literal(data=record_data, path=req_path)
            if isinstance(req_value, bool):
                if req_value is True:
                    if not record_value:
                        return False
                elif req_value is False:
                    if record_value:
                        return False
            elif record_value != req_value:
                return False

        return True


class OutputRow():
    def __init__(self, **kwargs):
        self.data = kwargs.get("data")
        self.pointer = kwargs.get("pointer")
        self.write_pointer = kwargs.get("write_pointer")
        self.pid = kwargs.get("record_pid")
        self.explode = {"explode_on": kwargs.get("explode_on"),
                        "explode_ordinal": kwargs.get("explode_ordinal"),
                        "explode_parent_value": kwargs.get("explode_parent_value")}
        self.group_role = kwargs.get("group_role")

        self.rules = kwargs.get("rules")
        self.write_out = True

        self.values = {}


# Static output functions
def remove_restricted_localities():
    # Get list of all event records with restricted locality
    if not read_config("localities_removed"):
        restricted_events = [i for i in memo if memo[i].get("restrict_locality")]
        for event_pid in restricted_events:
            remove_event_coordinates(event_pid)

            restricted_objects = list_restricted_objects(event_pid)
            if restricted_objects:
                remove_object_locality(restricted_objects)
                remove_object_images(restricted_objects)

        if not read_config("quiet"):
            print("Locality information removed from restricted specimens")
        write_config("localities_removed", True)


def remove_event_coordinates(event_pid):
    # Remove decimal latitude, longitude, and datum from mappingDetails
    loc_data = records.records[event_pid].data.get("atLocation")
    mapping_data = None
    if loc_data:
        try:
            mapping_data = loc_data["mappingDetails"][0]
        except (KeyError, IndexError):
            pass

    if mapping_data:
        if not read_config("quiet"):
            print("Removing locality information from {}".format(event_pid))
        records.records[event_pid].data["atLocation"]["mappingDetails"][0]["decimalLatitude"] = None
        records.records[event_pid].data["atLocation"]["mappingDetails"][0]["decimalLongitude"] = None
        records.records[event_pid].data["atLocation"]["mappingDetails"][0]["geodeticDatum"] = None


def list_restricted_objects(event_pid):
    # Create list of restricted objects from memo to allow redactions
    object_records = [i for i in memo if memo[i].get("associated_event") == event_pid]
    if len(object_records) > 0:
        return object_records
    else:
        return None


def remove_object_locality(restricted_objects):
    # Remove locality from objects associated with restricted events
    for record_pid in restricted_objects:
        if not read_config("quiet"):
            print("Removing locality information from related record {}".format(record_pid))
        records.records[record_pid].data["evidenceFor"]["atEvent"]["atLocation"]["locality"] = None
        records.records[record_pid].data["restrictLocalityData"] = True


def remove_object_images(restricted_objects):
    # Remove images from objects associated with restricted events
    # Keep images if they're of types
    for record_pid in restricted_objects:
        type_status = [i for i in collate_list(records.records[record_pid].data,
                                               ["identification", "i", "typeStatus"]) if i]
        if len(type_status) == 0:
            if not read_config("quiet"):
                print("Removing images from related record {}".format(record_pid))
            records.records[record_pid].data["hasRepresentation"] = None


def remove_nonwriting_images(record):
    # Removes images not being written out if only some selected
    memo_record = memo.get(record.pid)
    if memo_record:
        if len(memo_record["media_irns"]) > 0:
            if len(memo_record["media_irns"]) < len(record.data.get("hasRepresentation")):
                has_rep_section = record.data.get("hasRepresentation")
                new_has_rep_section = [sec for sec in has_rep_section if sec.get("id") in memo_record["media_irns"]]
                record.data["hasRepresentation"] = new_has_rep_section

    return record



class RecordStore():
    def __init__(self):
        self.records = {}

    def append(self, record):
        this_pid = record.pid
        self.records.update({this_pid: record})

    def find_record(self, endpoint, irn):
        this_pid = "tepapa:collection/{e}/{i}".format(e=endpoint, i=irn)
        return self.records.get(this_pid)


records = RecordStore()



from datetime import datetime

from src.setup.Settings import read_config
from src.monitoring.Stats import stats
from src.datastore.Memo import memo, format_pid, retrieve_from_memo, add_to_memo, update_memo


class Record():
    def __init__(self, data, endpoint):
        self.data = data
        self.endpoint = endpoint
        self.irn = self.data.get("id")
        self.pid = self.data.get("pid")

        if read_config("restrict_locality"):
            self.check_locality_restriction()

        if read_config("compare_quality_score"):
            self.check_quality_score()

        self.check_new_or_modified()

        if not read_config("quiet"):
            print("Record created: {e}, {i}".format(e=self.endpoint, i=self.irn))

    def check_locality_restriction(self):
        # If restrict locality is on, find event records that need to be redacted
        if self.endpoint == "object":
            try:
                event_id = self.data["evidenceFor"]["atEvent"]["id"]
            except KeyError:
                event_id = None

            if event_id:
                if self.data.get("restrictLocalityData"):
                    event_pid = add_to_memo(irn=event_id, endpoint="fieldcollection", extension=True)
                    update_memo(event_pid, "restrict_locality", True)

                event_pid = format_pid(endpoint="fieldcollection", irn=event_id)
                update_memo(self.pid, "associated_event", event_pid)

    def check_quality_score(self):
        # If part of the primary harvest, save the upper and lower bounds of qualityScores for the result set
        if self.endpoint == read_config("endpoint"):
            try:
                quality_score = self.data["_meta"]["qualityScore"]
            except KeyError:
                quality_score = None

            if quality_score:
                if not stats.quality_score_lower:
                    stats.quality_score_lower = quality_score
                else:
                    if quality_score < stats.quality_score_lower:
                        stats.quality_score_lower = quality_score
                if not stats.quality_score_upper:
                    stats.quality_score_upper = quality_score
                else:
                    if quality_score > stats.quality_score_upper:
                        stats.quality_score_upper = quality_score

    def check_new_or_modified(self):
        if not retrieve_from_memo(self.pid)["checked_if_modified"]:
            modified_recently = self.compare_dates("modified")
            created_recently = self.compare_dates("created")

            update_memo(self.pid, "checked_if_modified", True)
            update_memo(self.pid, "modified_recently", modified_recently)
            update_memo(self.pid, "created_recently", created_recently)

    def compare_dates(self, datetype):
        record_date = self.data["_meta"][datetype]
        rd_datestamp = datetime.fromisoformat(record_date.replace("Z", ""))
        if rd_datestamp >= read_config("check_modified_since"):
            return True
        else:
            return False



import click
from src.setup.Settings import (read_config, write_config, setup_project, update_query, add_limit_to_config,
                                update_settings, update_list_source)
from src.setup.InputList import read_input_list
from src.monitoring.Stats import stats
from src.util.ExportHandler import create_export_dir, delete_old_exports
from src.monitoring.ExportReport import generate_export_report
from src.setup.Mapper import Mapping
from src.util.Harvester import Harvester


# Harvests records from a search or list of IRNs
# Harvested data is processed according to rules supplied in map
# Processed data is written out to one or more files
def blasto():
    print("Looks like team blasto is blasting off again")
    delete_old_exports()

    # Start stats
    stats.mark_time("start")

    write_config("mapping", Mapping())
    harvester = Harvester()

    # Read mode from config and run required harvest
    # Passes harvested data directly to processing
    mode = read_config("mode")
    if mode == "list":
        read_input_list(read_config("list_source"))
        harvester.harvest_from_list()
    else:
        harvester.complete_harvest(mode)

    # (Hopefully) temporary workaround to some ETL coordinate transformation
    if read_config("coordinate_workaround"):
        from src.util.CoordinateWorkaround import apply_coordinate_workaround
        apply_coordinate_workaround()

    # Create export directory
    create_export_dir()

    # Write out harvested and processed data
    mapping = read_config("mapping")
    stats.mark_time("processing start")
    for output in mapping.outputs:
        output.write_output()
    stats.mark_time("processing end")

    # Stop stats and write results
    stats.mark_time("end")
    generate_export_report()


@click.command()
@click.option('--project', default='none', help='name of pre-defined project')
@click.option('--query', default='none', help='override query string in config')
@click.option('--listsource', default='none', help='override list source in config')
@click.option('--limit', default=-1, help='override max record limit in config when testing')
@click.option('--profiler', default=False, help='run with cProfile')
@click.option('--exportid', default=None, help='provide an id to use when writing out script results')
def cli(project, query, listsource, limit, profiler, exportid):
    setup_project(project)
    update_query(query)
    update_list_source(listsource)
    add_limit_to_config(limit)
    write_config("export_id", exportid)
    update_settings()

    if profiler:
        from src.monitoring import Profiler
        Profiler.profiler.enable()
        blasto()
        Profiler.profiler.disable()
        Profiler.output_profile()
    else:
        blasto()


if __name__ == "__main__":
    cli()


import os
import yaml
import json
from datetime import datetime, timedelta
from src.setup.Validator import ConfigValidator


config = {}
collections = ["Archaeozoology", "Art", "Birds", "CollectedArchives", "Crustacea", "Fish",
               "FossilVertebrates", "Geology", "History", "Insects", "LandMammals",
               "MarineInvertebrates", "MarineMammals", "Molluscs", "MuseumArchives",
               "PacificCultures", "Philatelic", "Photography", "Plants", "RareBooks",
               "ReptilesAndAmphibians", "TaongaMāori"]
humanities = ["Art", "CollectedArchives", "History", "MuseumArchives", "PacificCultures",
              "Philatelic", "Photography", "RareBooks", "TaongaMāori"]
sciences = ["Archaeozoology", "Birds", "Crustacea", "Fish", "FossilVertebrates", "Geology",
            "Insects", "LandMammals", "MarineInvertebrates", "MarineMammals", "Molluscs",
            "Plants", "ReptilesAndAmphibians"]
defaults = {"mapfile": "default.yaml",
            "endpoint": "object",
            "query": "*",
            "timeout": 5,
            "attempts": 3,
            "days_since_modified": 31}


def read_config(key):
    value = config.get(key)
    return value


def write_config(key, value):
    global config
    config[key] = value


def setup_project(project):
    if project != "none":
        config_file = "src/projects/{}_config.yaml".format(project)

    else:
        config_file = "config.yaml"

    if not os.path.exists(config_file):
        config_file = "src/projects/default.yaml"

    read_config_file(config_file, project)


def read_config_file(config_file, project):
    global config
    print(os.getcwd())
    try:
        with open(config_file, "r", encoding="utf-8") as f:
            print("Reading {}".format(config_file))
            config = yaml.safe_load(f)
    except IOError:
        break_on_settings_error("No config file found")

    config_validator = ConfigValidator(config)
    if config_validator.fail:
        break_on_settings_error("Config validation failed")

    if project != "none":
        write_config("project_name", project)


def update_query(query):
    if query != "none":
        write_config("query", query)


def update_list_source(list_source):
    if list_source != "none":
        write_config("list_source", list_source)


def add_limit_to_config(limit):
    if limit != -1:
        write_config("record_limit", limit)


def update_settings():
    set_mode()
    set_mapfile()
    set_input_dir()
    set_output_dir()
    set_skiplist()
    set_endpoint()
    load_country_codes()
    set_check_modified()
    format_export_string()

    # API parameters
    set_api_key()
    set_timeout()
    set_attempts()

    # List mode parameters
    set_list_source()

    # Search mode parameters
    set_query()
    if config.get("mode") != "list":
        set_filters()

    if not read_config("quiet"):
        print("Settings updated...")


def set_mode():
    # Return error if no mode selected
    mode = read_config("mode")
    if not mode:
        break_on_settings_error("No mode selected - must be search, scroll or list")
    else:
        if mode not in ["search", "scroll", "list"]:
            break_on_settings_error("Mode must be search, scroll or list")


def set_mapfile():
    # Defaults to defaultmap if no map provided
    map_file = read_config("mapping_file")
    if not map_file:
        map_file = defaults["mapfile"]
        print("No mapfile provided. Setting to default map")

    # Point to map
    map_path = "src/resources/mapfiles/{}".format(map_file)
    write_config("mapping_file", map_path)


def set_input_dir():
    # Return error if no input directory
    input_dir = read_config("input_dir")
    if not input_dir:
        break_on_settings_error("Input directory location needed")
    else:
        if not os.path.exists(input_dir):
            os.mkdir(input_dir)


def set_output_dir():
    # Return error if no output directory
    output_dir = read_config("output_dir")
    if not output_dir:
        break_on_settings_error("Output directory location needed to write files")
    else:
        if not os.path.exists(output_dir):
            os.mkdir(output_dir)


def set_skiplist():
    # Reads in list of IRNs to skip if provided
    if read_config("use_skiplist"):
        populate_skiplist()


def populate_skiplist():
    skiplist = []
    skipfile = "{d}/{f}".format(d=read_config("input_dir"), r=read_config("skipfile"))
    with open(read_config("skipfile"), 'r', encoding="utf-8") as f:
        lines = f.readlines()

    for line in lines:
        skiplist.append(int(line.strip()))

    write_config("skiplist", skiplist)

    if not read_config("quiet"):
        print("Skiplist populated")


def set_endpoint():
    # Default to object endpoint if none provided
    if not read_config("endpoint"):
        write_config("endpoint", defaults["endpoint"])
        print("No endpoint provided. Setting to object")


def load_country_codes():
    try:
        with open("src/resources/countrycodes.json") as f:
            country_codes = json.load(f)
            write_config("country_codes", country_codes)
    except IOError:
        if not read_config("quiet"):
            print("No country code file found")


def set_check_modified():
    days_since_modified = read_config("days_since_modified")
    if not days_since_modified:
        days_since_modified = defaults["days_since_modified"]
        write_config("days_since_modified", days_since_modified)

    check_modified_since = datetime.today() - timedelta(days=days_since_modified)
    write_config("check_modified_since", check_modified_since)


def format_export_string():
    project_name = read_config("project_name")
    export_id = read_config("export_id")
    export_filename_string = datetime.now().strftime("%Y%m%d-%H%M%S")
    if project_name:
        export_filename_string += "-{}".format(project_name)
    if export_id:
        export_filename_string += "-{}".format(export_id)

    write_config("export_filename", export_filename_string)


def set_api_key():
    # Return error if no API key
    if read_config("api_key_env"):
        try:
            write_config("api_key", os.environ[read_config("api_key_env")])
            if not read_config("quiet"):
                print("API key set to {}".format(read_config("api_key")))
        except ValueError:
            break_on_settings_error("No API key found in environment")
    else:
        break_on_settings_error("API key required to live")


def set_timeout():
    # Set timeout to 5 seconds if not provided
    if not read_config("timeout"):
        write_config("timeout", defaults["timeout"])


def set_attempts():
    # Set attempt count to 3 if not provided
    if not read_config("attempts"):
        write_config("attempts", defaults["attempts"])


def set_list_source():
    # Concatenate input directory and list source filename
    list_source = read_config("list_source")
    if list_source:
        list_filepath = "{d}/{f}".format(d=read_config("input_dir"), f=read_config("list_source"))
        write_config("list_source", list_filepath)


def set_query():
    # Set search query to wildcard if not provided
    if not read_config("query"):
        write_config("query", defaults["query"])


def set_filters():
    filters = []

    # Filter to a user-selected collection
    filter_config = read_config("filter")
    user_collection = filter_config.get("collection")
    if user_collection:
        user_collection = user_collection.split("|")
        colls = []
        coll_types = []
        for c in user_collection:
            if c in collections:
                colls.append(c)
                if user_collection in humanities:
                    coll_types.append("Object")
                elif user_collection in sciences:
                    coll_types.append("Specimen")

        if len(colls) > 0:
            if len(colls) == 1:
                coll_filter_value = colls[0]
            else:
                coll_filter_value = colls
            filters.append({"field": "collection", "keyword": coll_filter_value})

        if len(coll_types) > 0:
            coll_types = list(set(coll_types))
            if len(coll_types) == 1:
                coll_type_filter_value = coll_types[0]
            else:
                coll_type_filter_value = coll_types
            filters.append({"field": "type", "keyword": coll_type_filter_value})
    else:
        if not read_config("quiet"):
            print("No valid collection in filter config. Getting records for all collections")

    # Filter to records with downloadable images
    if filter_config.get("allows_download") is not None:
        if filter_config.get("allows_download"):
            filters.append({"field": "hasRepresentation.rights.allowsDownload", "keyword": "True"})
        elif not filter_config.get("allows_download"):
            filters.append({"field": "hasRepresentation.rights.allowsDownload", "keyword": "False"})

    # Filter using user-selected keywords
    kw_fields = filter_config.get("keyword_fields")
    kw_values = filter_config.get("keyword_values")

    if kw_fields:
        kw_fields = kw_fields.split("|")
        kw_values = kw_values.split("|")
        for i in range(0, len(kw_fields)):
            filters.append({"field": kw_fields[i], "keyword": kw_values[i]})

    write_config("filters", filters)


def break_on_settings_error(message):
    exit("Error during setup: {}".format(message))



import logging
# Read more about this
# https://www.loggly.com/ultimate-guide/python-logging-basics/
# https://realpython.com/python-logging/

logging.basicConfig(level=logging.DEBUG,
                    filename="src/logging/blastochor.log",
                    filemode="w",
                    format="%(asctime)s - %(name)s - %(process)d - %(levelname)s - %(message)s")

logging.debug("This is a debug message")
logging.info("This is an info message")
logging.warning("This is a warning message")
logging.error("This is an error message")
logging.critical("This is a critical message")


from src.setup.Settings import read_config
from src.util.ProcessingFunctions import ValueContainer, literal


class RowProcessor():
    def __init__(self, rules, output_row, requires):
        self.rules = rules
        self.output_row = output_row

        if requires:
            if not self.check_requirements(requires):
                self.output_row.write_out = False

        if self.output_row.write_out:
            self.run_field_processing()

    def check_requirements(self, requires):
        for requirement in requires:
            requirement_value = requirement["condition"]
            row_value = literal(data=self.output_row.data,
                                path=requirement["path"].split("."),
                                ordinal=self.output_row.explode.get("explode_ordinal"))
            if isinstance(requirement_value, bool):
                if requirement_value:
                    if not row_value:
                        return False
                elif not requirement_value:
                    if row_value:
                        return False

            elif row_value != requirement_value:
                return False

        return True

    def run_field_processing(self):
        for rule in self.rules:
            value = None
            fieldname = rule.output_fieldname
            if read_config("group_rows"):
                if self.check_for_inclusion(fieldname):
                    value = ValueContainer(rule=rule, output_row=self.output_row).current_value
            else:
                value = ValueContainer(rule=rule, output_row=self.output_row).current_value

            if not value and (value != 0):
                value = ""

            self.output_row.values.update({fieldname: value})

    def check_for_inclusion(self, fieldname):
        # If grouping exploded rows, check if specific fields should be included/excluded
        row_role = self.output_row.group_role
        match row_role:
            case "child":
                fields = "child_fields"
            case "parent":
                fields = "parent_fields"
            case _:
                fields = "ungrouped_fields"

        if include_field(fieldname, read_config(fields)):
            return True


def include_field(fieldname, fields):
    include = fields.get("include")
    column_names = fields.get("fields").split(", ")

    if include:
        if fieldname in column_names:
            return True
    elif not include:
        if fieldname not in column_names:
            return True



memo = {}


def format_pid(endpoint="object", irn=None):
    if irn:
        return "tepapa:collection/{e}/{i}".format(e=endpoint, i=irn)
    else:
        return None


def retrieve_from_memo(pid=None):
    memo_record = memo.get(pid)
    return memo_record


def add_to_memo(status=None, irn=None, endpoint=None, label=None, extension=None):
    global memo
    pid = format_pid(endpoint=endpoint, irn=irn)
    memo_record = retrieve_from_memo(pid)
    if memo_record:
        return memo_record["pid"]
    else:
        if not status:
            status = "pending"
        if not extension:
            extension = False

        memo[pid] = {
            "status": status,
            "irn": irn,
            "endpoint": endpoint,
            "pid": pid,
            "write_to": [],
            "structure": {},
            "is_extension": extension,
            "media_irns": [],
            "checked_if_modified": False,
            "created_recently": False,
            "modified_recently": False
        }

        if label:
            memo[pid]["write_to"].append(label)
            memo[pid]["structure"].update({label: {"write": True, "extends": []}})

        return pid


def update_memo(pid, update_type, value, label=None):
    global memo
    if retrieve_from_memo(pid):
        match update_type:
            case "add_label":
                if value not in memo[pid]["write_to"]:
                    memo[pid]["write_to"].append(value)
                    memo[pid]["structure"].update({value: {"write": True, "extends": []}})

            case "extends":
                if label:
                    if value not in memo[pid]["structure"][label]["extends"]:
                        memo[pid]["structure"][label]["extends"].append(value)
                memo[pid]["is_extension"] = True

            case "media":
                memo[pid]["media_irns"].append(value)

            case _:
                memo[pid][update_type] = value
    else:
        return False



from datetime import datetime
from math import floor
from src.setup.Settings import read_config
from src.datastore.Memo import memo


class ApplicationStats():
    def __init__(self):
        self.app_start = None
        self.app_end = None
        self.harvest_start = None
        self.harvest_end = None
        self.extension_records_start = None
        self.extension_records_end = None
        self.processing_start = None
        self.processing_end = None
        self.writing_start = None
        self.writing_end = None

        self.run_time = None
        self.harvest_time = None
        self.extension_time = None
        self.processing_time = None
        self.writing_time = None

        self.api_call_count = 0
        self.initial_record_count = 0
        self.source_list_count = 0
        self.extension_records_count = 0
        self.new_record_count = None
        self.modified_record_count = None

        self.file_write_counts = {}
        self.export_filenames = []

        self.quality_score_lower = None
        self.quality_score_upper = None

    def mark_time(self, task):
        timestamp = datetime.now()
        match task:
            case "start":
                self.app_start = timestamp
            case "end":
                self.app_end = timestamp
            case "harvest start":
                self.harvest_start = timestamp
            case "harvest end":
                self.harvest_end = timestamp
            case "extension start":
                self.extension_records_start = timestamp
            case "extension end":
                self.extension_records_end = timestamp
            case "processing start":
                self.processing_start = timestamp
            case "processing end":
                self.processing_end = timestamp

    def process_runtimes(self):
        # Record how long script took to run in total
        delta = self.app_end - self.app_start
        self.run_time = datetime_to_string(delta.total_seconds())

        # Record how long script took to perform initial harvest
        delta = self.harvest_end - self.harvest_start
        self.harvest_time = datetime_to_string(delta.total_seconds())

        # Record how long script took to perform extra queries
        delta = self.extension_records_end - self.extension_records_start
        self.extension_time = datetime_to_string(delta.total_seconds())

        # Record how long script took to perform processing
        delta = self.processing_end - self.processing_start
        self.processing_time = datetime_to_string(delta.total_seconds())

    def count_new_and_modified(self):
        self.new_record_count = self.count_modified_records("created")
        self.modified_record_count = self.count_modified_records("modified")

    def count_modified_records(self, modificationtype):
        modified_records = [memo[i] for i in list(memo.keys()) if memo[i]["{}_recently".format(modificationtype)]]
        object_count = len([i for i in modified_records if i["endpoint"] == "object"])
        agent_count = len([i for i in modified_records if i["endpoint"] == "agent"])
        taxon_count = len([i for i in modified_records if i["endpoint"] == "taxon"])
        return {"object": object_count,
                "agent": agent_count,
                "taxon": taxon_count}

    def print_stats(self):
        print("Script ran in {}".format(self.run_time))
        print("Script made {} API calls".format(self.api_call_count))

        if read_config("mode") != "list":
            print("Search returned {} results".format(self.initial_record_count))
        else:
            print("Source list contained {} records".format(self.source_list_count))
        print("Harvesting ran in {}".format(self.harvest_time))

        print("Queried {} extension records".format(self.extension_records_count))
        print("Getting extra records took {}".format(self.extension_time))

        print("Processing records took {}".format(self.processing_time))

        for label in self.file_write_counts.keys():
            print("Wrote {n} records to the {l} file".format(n=self.file_write_counts.get(label), l=label))

        print("Export analysis:")
        self.print_counts(self.new_record_count, "new")
        self.print_counts(self.modified_record_count, "updated")

    def print_counts(self, modified_counts, t):
        days = read_config("days_since_modified")
        for endpoint, count in modified_counts.items():
            count_string = "{n} {e} records retrieved {t} in the last {x} days".format(n=count,
                                                                                       e=endpoint,
                                                                                       t=t,
                                                                                       x=days)
            print("\t{}".format(count_string))


def datetime_to_string(seconds):
    if seconds > 3600:
        hours = seconds // 3600
        minutes = (seconds // 60) % 60
        return "{h} hours and {m} minutes".format(h=floor(hours), m=floor(minutes))
    elif seconds > 60:
        minutes = seconds // 60
        seconds = seconds - (minutes * 60)
        return "{m} minutes and {s} seconds".format(m=floor(minutes), s=floor(seconds))
    else:
        return "{} seconds".format(round(seconds))


stats = ApplicationStats()



from math import floor
import random
import nh3
from askCO import Resource
from geopy.distance import geodesic as GD
from src.setup.Settings import read_config
from src.monitoring.Stats import stats
from src.datastore.RecordStore import records


class ValueContainer():
    def __init__(self, output_row, rule):
        self.output_row = output_row
        self.rule = rule
        self.current_value = None
        self.reprocess = False
        self.complete = False

        self.run_functions()

    def run_functions(self):
        for function in self.rule.mapping_functions:
            if not self.complete:
                container = FunctionContainer(self.output_row,
                                              self.current_value,
                                              self.reprocess,
                                              function,
                                              None)
                self.current_value = container.current_value
                self.reprocess = container.reprocess
                self.complete = container.complete

        self.complete = True


class FunctionContainer():
    def __init__(self, output_row=None, current_value=None, reprocess=None, function=None, value_index=None):
        self.output_row = output_row
        self.current_value = current_value
        self.reprocess = reprocess
        self.function = function
        self.value_index = value_index
        self.complete = False

        self.unpack_function()

    def unpack_function(self):
        function_name = list(self.function.keys())[0]
        params = self.function[function_name]

        if not read_config("quiet"):
            print("Running function: {}".format(function_name))

        processed_function = FunctionProcessor(function_name,
                                               params,
                                               self.output_row,
                                               self.current_value,
                                               self.reprocess,
                                               self.value_index)

        self.current_value = processed_function.output_value
        self.reprocess = processed_function.reprocess
        self.complete = processed_function.complete


class FunctionProcessor:
    def __init__(self, function=None, params=None, output_row=None, input_value=None, reprocess=False, value_index=None):
        self.function = function
        self.params = params
        self.output_row = output_row
        self.data = self.output_row.data

        if not reprocess:
            self.input_value = self.data
        else:
            self.input_value = input_value

        self.output_value = None
        self.reprocess = True
        self.complete = False

        self.value_index = value_index

        self.select_function()

    def select_function(self):
        match self.function:
            case "continue_if":
                if literal(data=self.input_value, path=self.params["path"]) != self.params["condition"]:
                    self.complete = True
                self.reprocess = False

            case "for_each":
                if isinstance(self.input_value, list):
                    self.output_value = []
                    value = None
                    value_index = 0
                    for value in self.input_value:
                        subfunction_reprocess = True
                        for subfunction in self.params:
                            subfunction_container = FunctionContainer(self.output_row,
                                                                      value,
                                                                      subfunction_reprocess,
                                                                      subfunction,
                                                                      value_index)
                            value = subfunction_container.current_value
                            subfunction_reprocess = subfunction_container.reprocess
                            subfunction_complete = subfunction_container.complete

                            if isinstance(value, list):
                                value = concatenate(value)

                        self.output_value.append(value)
                        value_index += 1

            case "fallback":
                if self.input_value is not None or (self.input_value == 0):
                    self.output_value = self.input_value
                else:
                    subfunction_reprocess = False
                    value = None
                    for subfunction in self.params:
                        subfunction_container = FunctionContainer(self.output_row,
                                                                  value,
                                                                  subfunction_reprocess,
                                                                  subfunction,
                                                                  None)
                        value = subfunction_container.current_value
                        subfunction_reprocess = subfunction_container.reprocess
                        subfunction_complete = subfunction_container.complete

                    self.output_value = value

            case "clean_html":
                self.output_value = clean_html(data=self.input_value)

            case "collate_list":
                path_param = self.params["path"]
                if "," in path_param:
                    self.output_value = []
                    paths = path_param.split(", ")
                    for path in paths:
                        path_values = collate_list(data=self.input_value,
                                                   path=path)
                        if path_values:
                            self.output_value.extend(path_values)
                else:
                    self.output_value = collate_list(data=self.input_value,
                                                     path=path_param)

            case "concatenate":
                self.output_value = concatenate(data=self.input_value)

            case "conditional":
                value_path = self.params["path"].split(".")
                condition_path = self.params["condition_path"].split(".")
                condition_values = self.params["match"].split("|")
                fallback_values = self.params.get("fallback_match")
                self.output_value = conditional_inclusion(data=self.input_value,
                                                          value_path=value_path,
                                                          condition_path=condition_path,
                                                          condition_values=condition_values,
                                                          fallback_values=fallback_values)

            case "country_code":
                path = self.params["path"]
                self.output_value = country_code(data=self.input_value,
                                                 path=path)

            case "create_filename":
                suffix = self.params["suffix"]
                self.output_value = create_filename(data=self.input_value,
                                                    suffix=suffix)

            case "first_match":
                paths = self.params["path"].split(", ")
                self.output_value = first_match(data=self.input_value,
                                                paths=paths)

            case "format_string":
                string = self.params["string"]
                paths = self.params["path"].split(", ")
                required = self.params.get("required")
                strip = self.params.get("strip")
                ordinal = self.output_row.explode.get("explode_ordinal")
                record_pid = self.output_row.pid
                self.output_value = format_string(data=self.input_value,
                                                  string=string,
                                                  paths=paths,
                                                  required=required,
                                                  strip=strip,
                                                  ordinal=ordinal,
                                                  record_pid=record_pid)

            case "hardcoded":
                value = self.params["value"]
                if value == "ordinal":
                    modifier = self.params.get("ordinal_modifier")
                    if self.output_row.explode.get("explode_ordinal"):
                        value = int(self.output_row.explode.get("explode_ordinal"))
                    else:
                        value = 0
                    if modifier:
                        value += modifier
                self.output_value = hardcode_value(value)

            case "identification_qualifier":
                taxon_path = self.params["taxon_path"]
                qualifier_path = self.params["qualifier_path"]
                ordinal = self.params.get("ordinal")
                if not ordinal and (ordinal != 0):
                    ordinal = self.output_row.explode.get("explode_ordinal")
                self.output_value = format_identification_qualifier(data=self.input_value,
                                                                    taxon_path=taxon_path,
                                                                    qualifier_path=qualifier_path,
                                                                    ordinal=ordinal)

            case "literal":
                path = self.params["path"].split(".")
                if self.output_row.explode.get("explode_on") in path:
                    ordinal = self.output_row.explode.get("explode_ordinal")
                elif self.value_index is not None or (self.value_index == 0):
                    ordinal = self.value_index
                else:
                    ordinal = self.params.get("ordinal")
                self.output_value = literal(data=self.input_value,
                                            path=path,
                                            ordinal=ordinal)

            case "lookup":
                endpoint = self.params["endpoint"]
                lookup_record = records.find_record(irn=self.input_value,
                                                    endpoint=endpoint)
                if lookup_record:
                    self.output_value = lookup_record.data
                else:
                    self.output_value = None

            case "map_value":
                self.output_value = map_from_value(data=self.input_value,
                                                   mappings=self.params)

            case "measure_geodistance":
                units = self.params["units"]
                self.output_value = measure_geodistance(data=self.input_value,
                                                        units=units)

            case "must_match":
                authorities = [str(term).lower() for term in self.params["terms"].split(", ")]
                self.output_value = must_match(data=self.input_value,
                                               authorities=authorities)

            case "related":
                size = self.params["size"]
                types = self.params["types"]
                self.output_value = get_related(data=self.input_value,
                                                size=size,
                                                types=types)

            case "set_priority":
                self.output_value = process_quality_score(data=self.input_value)

            case "truncate":
                length = self.params["length"]
                suffix = self.params.get("suffix")
                self.output_value = truncate_value(data=self.input_value,
                                                   length=length,
                                                   suffix=suffix)

            case "use_config":
                key = self.params["key"]
                self.output_value = read_config(key)

            case "use_group_labels":
                label = None
                if self.output_row.group_role == "parent":
                    label = self.params["parent"]
                elif self.output_row.group_role == "child":
                    label = self.params["child"]
                elif not self.output_row.group_role:
                    label = self.params["other"]
                self.output_value = label


def split_path(path):
    # Ensure paths are lists that can be iterated through
    if not isinstance(path, list):
        path = path.split(".")
    return path


def handle_iterator_in_path(path=None, ordinal=None):
    # Replace the list indicator 'i' in a path with a specified ordinal to create a navigable path
    # If no ordinal provided, defaults to first item in the list
    if isinstance(path, list):
        if "i" in path:
            if not ordinal:
                ordinal = 0

            path = [ordinal if item == "i" else item for item in path]

    return path


def step_to_field(data=None, path=None):
    # Step through a record to provide a subsection or value
    if data:
        if isinstance(path, str):
            return data.get(path)
        else:
            step = data
            for field in path:
                try:
                    step = step[field]
                except (KeyError, IndexError):
                    return None
            return step
    else:
        return None


def clean_html(data=None):
    if data:
        value = nh3.clean(data)
    else:
        value = None

    return value


def collate_list(data=None, path=None):
    # Look through each child path of a list and collate all values of child fields with the same name
    # Currently only works for one level of list
    path = split_path(path)
    values = []

    parent_path = []
    for field in path:
        if field == "i":
            break
        else:
            parent_path.append(field)

    parent_data = step_to_field(data, parent_path)
    if not parent_data:
        return None

    # Put a limit on the number of list items returned
    # Long lists (eg people referenced by a lot of narratives) can break the CSV cell limit
    iterator_limit = len(parent_data)
    app_limit = read_config("max_list_size")
    if app_limit:
        if 0 < app_limit < iterator_limit:
            iterator_limit = app_limit

    for i in range(0, iterator_limit):
        value = literal(data, path=path, ordinal=i)
        values.append(value)

    if len(values) == 0:
        values = None

    return values


def concatenate(data=None):
    # Join a list using "|" as specified in Darwin Core
    if isinstance(data, list):
        data = [str(i) for i in data if i is not None]
        if len(data) == 1:
            value = data[0]
        elif len(data) > 1:
            value = " | ".join(data)
        else:
            value = None
        return value
    else:
        return data


def conditional_inclusion(data, value_path, condition_path, condition_values, fallback_values):
    # Return a value if another field matches a specified value
    # Both paths must be in the same section of the record
    # For example, find related.i.contentUrl if related.i.title is "ORCID"
    if "i" in condition_path:
        list_to_check = collate_list(data, path=condition_path)
        if not list_to_check:
            return None
        else:
            found_values = []
            returned_value_index = 0
            for returned_value in list_to_check:
                if returned_value in condition_values:
                    found_values.append(literal(data, path=value_path, ordinal=returned_value_index))
                returned_value_index += 1
            if len(found_values) == 0:
                if fallback_values:
                    fallback_values = fallback_values.split("|")
                    returned_value_index = 0
                    for returned_value in list_to_check:
                        if returned_value in fallback_values:
                            found_values.append(literal(data, path=value_path, ordinal=returned_value_index))
                        returned_value_index += 1
    else:
        returned_value = literal(data, path=condition_path)
        if not returned_value:
            return None
        else:
            found_values = []
            if returned_value in condition_values:
                found_values.append(literal(data, path=value_path))
            if len(found_values) == 0:
                if fallback_values:
                    fallback_values = fallback_values.split("|")
                    if returned_value in fallback_values:
                        found_values.append(literal(data, path=value_path))

    # Return a single value if there is only one match, a list if there are multiple, None if there aren't any
    if len(found_values) == 0:
        return None
    elif len(found_values) == 1:
        return found_values[0]
    else:
        return found_values


def country_code(data, path):
    country = literal(data, path)
    if country:
        country_lookup = next(filter(lambda country_dict: country_dict.get("Name") == country,
                                     read_config("country_codes")), None)
        if country_lookup:
            return country_lookup.get("Code")

    return None


def create_filename(data, suffix):
    # Remove unsafe characters from a string to form a usable filename
    data = data.replace(" ", "_")
    data = data.replace("?", "")
    data = data.replace("\"", "")
    data = data.replace(":", "")
    data = data.replace(";", "")
    data = data.replace(".", "")
    data = data.replace(",", "")
    data = data.replace("#", "")
    data = data.replace("*", "")
    data = data.replace("\'", "")
    data = data.replace("\\", "")
    data = data.replace("/", "")

    return "{d}.{s}".format(d=data, s=suffix)


def extend_value(data, length, string):
    # Extend a value to a specified length using a provided string
    if len(data) < length:
        return data + string


def first_match(data, paths):
    value = None
    for path in paths:
        path = split_path(path)
        if "i" in path:
            value = collate_list(data, path)
        else:
            value = literal(data, path)
        if value:
            return value

    return None


def format_identification_qualifier(data, taxon_path, qualifier_path, ordinal):
    # Special formatter for GBIF export (identificationQualifier field)
    taxon_id = literal(data, path=taxon_path, ordinal=ordinal)
    if taxon_id:
        taxon_record = records.find_record(endpoint="taxon", irn=taxon_id)
        if taxon_record:
            fields = ["subspecies", "species", "genus"]
            name = None
            for field in fields:
                if taxon_record.data.get(field):
                    name = taxon_record.data.get(field)
                    break

            if name:
                qualifier = literal(data, path=qualifier_path, ordinal=ordinal)
                if qualifier:
                    if qualifier in ["?", "s.l.", "s.s.", "pro. hyb.", "pro. sp.", "cv.", "sp. nov.", "sp. nov.?",
                                     "prob."]:
                        return "{n} {q}".format(n=name, q=qualifier)
                    elif qualifier in ["aff.", "cf."]:
                        return "{q} {n}".format(q=qualifier, n=name)
        else:
            if not read_config("quiet"):
                print("No record found for taxon record {}".format(taxon_id))

    return None


def format_string(data, string=None, paths=None, required=False, strip=None, ordinal=None, record_pid=None):
    # Find a value and incorporate it into a specified string
    values = []
    for path in paths:
        if path == "parent_id":
            if record_pid:
                record_id = record_pid.split("/")[-1]
                values.append(record_id)
            else:
                values.append("")
        elif path == "explode_ordinal":
            # Used for Google Arts orderid (index starts at 1)
            values.append(ordinal + 1)
        else:
            value = literal(data=data, path=path, ordinal=ordinal)
            if not value:
                if required:
                    continue
                else:
                    value = ""
            values.append(value)

    if required:
        spaces = string.count("{}")
        if len(values) != spaces:
            return None

    if strip:
        values = [value.replace(strip, "") for value in values]

    return string.format(*values)


def get_related(data, size, types):
    # Make a fresh query to the special /related endpoint for the current record
    endpoint = data.get("href").split("/")[-2]
    irn = data.get("id")

    query = Resource(
        quiet=read_config("quiet"),
        api_key=read_config("api_key"),
        timeout=read_config("timeout"),
        attempts=read_config("attempts"),
        sleep=0.1,
        endpoint=endpoint,
        irn=irn,
        related=True,
        size=size,
        types=types)

    query.send_query()
    stats.api_call_count += 1

    return query.records


def hardcode_value(value):
    # Return a specified string
    return value


def literal(data, path, ordinal=None):
    # Find a given path in a record and return the literal value
    path = split_path(path)
    path = handle_iterator_in_path(path=path, ordinal=ordinal)
    value = step_to_field(data=data, path=path)
    return value


def map_from_value(data, mappings):
    # If a specified value is present, map to a provided string
    if isinstance(data, list):
        mapped_value = []
        for value in data:
            mapped_value.append(mappings.get(value))
    else:
        mapped_value = mappings.get(data)

    return mapped_value


def measure_geodistance(data, units):
    # If a collection event covers a coordinate line, compares the centroid and start to find a radial distance
    centroid, radial = None, None
    if literal(data, "atLocation"):
        try:
            centroid_vals = literal(data, ["atLocation", "mappingCentroid"])
            if centroid_vals:
                centroid = (centroid_vals["lat"], centroid_vals["lon"])
        except KeyError:
            pass

        try:
            point_vals = literal(data, ["atLocation", "mappingDetails"])
            if point_vals:
                if len(point_vals) == 2:
                    radial = (point_vals[0]["decimalLatitude"], point_vals[0]["decimalLongitude"])
        except (KeyError, IndexError):
            pass

    if centroid and radial:
        radius_val = GD(centroid, radial).km
        if units == "m":
            radius_val = radius_val * 1000
        radius_int = floor(radius_val)
        return radius_int


def must_match(data, authorities):
    # Include value if it also appears in a provided list of terms
    matching = None
    if isinstance(data, list):
        matching = [term for term in data if str(term).lower() in authorities]
    else:
        if str(data).lower() in authorities:
            matching = data

    return matching


def process_quality_score(data):
    # Turns a record's qualityScore into an integer from 1-500
    # Used to assign priority in Google Arts & Culture
    priority = None

    if stats.quality_score_lower and stats.quality_score_upper:
        try:
            quality_score = data["_meta"]["qualityScore"]
            score = quality_score - stats.quality_score_lower
            upper = stats.quality_score_upper - stats.quality_score_lower

            score_percentage = score / upper * 100
            score_percentage = 100 - score_percentage

            priority = floor(score_percentage * 3)
            modifier = random.randint(-15, 15)
            priority = priority + modifier

            if priority <= 0:
                priority = random.randint(1, 200)

        except KeyError:
            pass

    return priority


def truncate_value(data, length, suffix):
    # Truncate a value to a specified length, with an optional suffix (eg "...")
    if isinstance(data, str):
        if len(data) > length:
            if suffix:
                suffix_length = len(suffix)
                string_length = length - suffix_length
                value = data[:string_length]
                value += suffix
            else:
                value = data[:length]

            return value

        else:
            return data

    else:
        return None



import cProfile
import pstats
from datetime import datetime

profiler = cProfile.Profile()


def output_profile():
    profile_stats = pstats.Stats(profiler)
    profile_stats.strip_dirs()
    now = datetime.now().strftime("%Y%m%d-%H_%M_%S")
    profile_stats.dump_stats("src/logging/{}.prof".format(now))



import csv
from src.setup.Settings import read_config
from src.monitoring.Stats import stats


class WriteFile():
    def __init__(self, label):
        self.label = label
        self.filename = self.generate_filename()

        self.file = open(self.filename, "w", newline="", encoding="utf-8")
        self.writer = csv.writer(self.file, delimiter=",")

    def generate_filename(self):
        output_dir = read_config("output_dir")
        export_dir = read_config("export_filename")
        write_filename = "{}-export.csv".format(self.label)
        stats.export_filenames.append(write_filename)
        write_location = "{d}/{e}/{f}".format(d=output_dir,
                                              e=export_dir,
                                              f=write_filename)
        return write_location

    def write_header_row(self, fieldnames):
        self.writer.writerow(fieldnames)

    def write_rows(self, output_rows, fieldnames):
        for row in output_rows:
            if row.write_out:
                self.write_this_row(row, fieldnames)

    def write_this_row(self, row, fieldnames):
        write_values = []
        if row.write_pointer:
            write_values.append(row.pointer)
        for field in fieldnames:
            value = row.values.get(field)
            # If a value is still a list after processing, turn it into a string
            if isinstance(value, list):
                value = [str(i) for i in value if i is not None]
                if len(value) == 0:
                    value = None
                elif len(value) == 1:
                    value = value[0]
                elif len(value) > 1:
                    value = " | ".join(value)
                else:
                    value = None
            # If a value is None, turn it into a blank string
            if not value and (value != 0):
                value = ""

            # If removing newlines is turned on, check if value is a string and replace them with spaces
            if read_config("clean_newlines"):
                if isinstance(value, str):
                    value = value.replace("\n", " ")

            write_values.append(value)

        self.writer.writerow(write_values)
        stats.file_write_counts[self.label] += 1



from src.setup.Settings import read_config
from src.datastore.RecordStore import records
from src.datastore.Memo import retrieve_from_memo
from lxml import etree


# When triggered, finds the EMu coordinate and datum values for collection event records
# and replaces the API values, avoiding the transformation applied in some cases by the ETL
def apply_coordinate_workaround():
    print("Starting coordinate workaround")
    emu_data_file = "{d}/resources/{f}".format(d=read_config("input_dir"),
                                               f=read_config("emufile"))

    # Define the table name. Tuples are the start of a new record if their parent is this
    table_name = "ecollectionevents"

    # Main loop to iterate through our records
    # We do this in chunks as the xml file can be quite large
    record_context = etree.iterparse(emu_data_file, events=('end',), tag='tuple')

    for event, record in record_context:
        # Check if the chunk is a record
        if not check_is_record(record, table_name):
            continue

        # Get the record's IRN or continue to next if no IRN
        event_irn = check_has_irn(record)
        if not event_irn:
            continue

        # Format pid from irn to allow lookup in memo
        event_pid = "tepapa:collection/fieldcollection/{}".format(event_irn)

        if not retrieve_from_memo(event_pid):
            continue

        # Get EMu values for latitude, longitude and datum
        lat_value = get_emu_value(record, "LatLatitudeDecimal")
        long_value = get_emu_value(record, "LatLongitudeDecimal")
        datum_value = get_emu_value(record, "LatDatum")

        # Map datum to the right EPSG code
        if datum_value:
            datum_value = map_datum_value(datum_value)

        # Update the saved record with the original EMu values
        update_record(event_pid, lat_value, long_value, datum_value)

    print("Coordinate workaround finished")


def check_is_record(record, table_name):
    parent = record.getparent()
    if parent is not None:
        parent_name = parent.attrib.get("name")
        if parent_name == table_name:
            return True
        else:
            return False
    else:
        return False


def check_has_irn(record):
    elem = record.find("atom[@name='irn']")
    if elem is not None and elem.text:
        irn = elem.text
        return irn
    else:
        return False


def get_emu_value(record, fieldname):
    values = []
    for atom in record.iter("atom"):
        atom_text = atom.text
        atom_name = atom.get("name")

        if atom is not None and atom_name == fieldname:
            values.append(atom_text)

    if len(values) > 0:
        # Records may have multiple sets of values but GBIF only uses the first
        return values[0]
    else:
        return None


def map_datum_value(datum_value):
    if "NZGD1949" in datum_value:
        return "EPSG:4272"
    elif "UTM" in datum_value:
        # Waiting on confirmation of code
        return "EPSG:4272"
    elif "NZGD2000" in datum_value:
        # Correct EPSG code TBD
        return "EPSG:4959"
    elif ("WGS1984" or "WGS84") in datum_value:
        return "WGS1984"


def update_record(event_pid, lat_value, long_value, datum_value):
    if records.records[event_pid].data["atLocation"].get("mappingDetails"):
        if lat_value:
            records.records[event_pid].data["atLocation"]["mappingDetails"][0]["decimalLatitude"] = lat_value
        if long_value:
            records.records[event_pid].data["atLocation"]["mappingDetails"][0]["decimalLongitude"] = long_value
        if datum_value:
            records.records[event_pid].data["atLocation"]["mappingDetails"][0]["geodeticDatum"] = datum_value



import yaml


class ConfigValidator():
    def __init__(self, config):
        self.config = config
        self.validation = load_validation("config")
        self.quiet = self.config.get("quiet")
        self.fail = False
        self.report = ["Validation report for config"]

        self.validate_keys()

        print_report("config", self.report, self.fail, self.quiet)

    def validate_keys(self):
        for key in self.validation["required"]:
            self.validate_key(key, "config", required=True)
        for key in self.validation["optional"]:
            self.validate_key(key, "optional")
        if self.config["group_rows"]:
            for key in self.validation["group"]:
                self.validate_key(key, "group", required=True)
        for key in self.validation["mode"][self.config["mode"]]:
            self.validate_key(key["key"], "mode", required=key["required"])
        for key in self.validation["filter"]:
            self.validate_key(key, "filter", parent="filter")

    def validate_key(self, key, key_type, required=False, parent=None):
        # Check if the config's required keys are present
        # Note this does not check the values of those keys
        if parent:
            check_list = self.config[parent]
        else:
            check_list = list(self.config.keys())
        present = False
        if key in check_list:
            present = True
        self.save_validation(key, key_type, required, present)

    def save_validation(self, key, key_type, required, present):
        message = None
        if present:
            if required:
                message = "Required {kt} key found: {k}".format(kt=key_type, k=key)
            else:
                message = "{kt} key found: {k}".format(kt=key_type.capitalize(), k=key)
        else:
            if required:
                self.fail = True
                message = "Missing required {kt} key: {k} not in config".format(kt=key_type, k=key)
            else:
                message = "Missing {kt} key: {k} not in config".format(kt=key_type.capitalize(), k=key)
        if message:
            self.report.append(message)


class MapValidator():
    def __init__(self, map_data, quiet, country_codes):
        self.map_data = map_data
        self.label = None
        self.validation = load_validation("map")
        self.functions = load_validation("functions")
        self.quiet = quiet
        self.country_codes = country_codes
        self.fail = False
        self.report = []

        self.validate_map()

        print_report("map", self.report, self.fail, self.quiet, self.label)

    def validate_map(self):
        if self.map_data.get("label"):
            self.label = self.map_data["label"]
            self.report.append("Validation report for map {}".format(self.map_data["label"]))
        else:
            self.fail = True
            self.report.append("Error: No label for this map.")

        for key in self.validation["required"]:
            self.validate_key(key, "map", required=True)
        for key in self.validation["optional"]:
            self.validate_key(key, "map")
        for key in self.validation["extension_keys"]:
            if self.map_data.get("extend"):
                for extension in self.map_data["extend"]:
                    self.validate_key(key, "extension", required=True, parent=extension)
        for key in self.validation["requires_keys"]:
            if self.map_data.get("requires"):
                for requirement in self.map_data["requires"]:
                    self.validate_key(key, "requires", required=True, parent=requirement)

        self.validate_fields()

    def validate_key(self, key, key_type, required=False, parent=None):
        # Check if the config's required keys are present
        # Note this does not check the values of those keys
        if parent:
            check_list = list(parent.keys())
        else:
            check_list = list(self.map_data.keys())
        present = False
        if key in check_list:
            present = True
        self.save_validation(key, key_type, required, present)

    def save_validation(self, key, key_type, required, present):
        message = None
        if present:
            if required:
                message = "Required {kt} key found: {k}".format(kt=key_type, k=key)
            else:
                message = "{kt} key found: {k}".format(kt=key_type.capitalize(), k=key)
        else:
            if required:
                self.fail = True
                message = "Missing required {kt} key: {k} not in map".format(kt=key_type, k=key)
            else:
                message = "Missing {kt} key: {k} not in map".format(kt=key_type.capitalize(), k=key)
        if message:
            self.report.append(message)

    def validate_fields(self):
        # Check each field against validation
        field_index = 1
        for field in self.map_data["fields"]:
            field_name = list(field.keys())[0]
            for function in field[field_name]:
                self.validate_function(field_name, field_index, function)
            field_index += 1

    def validate_function(self, field_name, field_index, function):
        function_name = list(function.keys())[0]
        # Check that the field's functions are valid and has suitable parameters
        if function_name == "for_each":
            message = "Field {i} ({n}): Function for_each contains subfunctions".format(i=field_index,
                                                                                        n=field_name)
            self.report.append(message)
            for subfunction in function[function_name]:
                self.validate_function(field_name, field_index, subfunction)
        elif function_name == "fallback":
            message = "Field {i} ({n}): Function fallback contains fallback functions".format(i=field_index,
                                                                                              n=field_name)
            self.report.append(message)
            for fallback_function in function[function_name]:
                self.validate_function(field_name, field_index, fallback_function)
        else:
            if function_name in self.functions:
                # Country code function requires file to be loaded
                if function_name == "country_code":
                    if not self.country_codes:
                        self.fail = True
                        message = "Field {i} ({n}): Required country codes have not been loaded".format(i=field_index,
                                                                                                        n=field_name)
                        self.report.append(message)

                valid_params = self.functions[function_name]["parameters"]
                if function.get(function_name):
                    field_params = list(function[function_name].keys())
                    for param_key in field_params:
                        self.validate_parameter(field_name,
                                                field_index,
                                                function_name,
                                                valid_params,
                                                param_key)
                    if valid_params:
                        for valid_param in valid_params:
                            if valid_param.get("required"):
                                self.check_for_required_params(field_name,
                                                               field_index,
                                                               function_name,
                                                               valid_param,
                                                               field_params)
                else:
                    self.report.append("Field {i} ({n}): Function has no parameters".format(i=field_index,
                                                                                            n=field_name))

            else:
                self.fail = True
                self.report.append("Field {i} ({n}): Function {f} is not a valid function".format(i=field_index,
                                                                                                  n=field_name,
                                                                                                  f=function_name))

    def validate_parameter(self, field_name, field_index, function_name, valid_params, param_key):
        if valid_params:
            if param_key in [i["parameter"] for i in valid_params]:
                # Decide whether to check if the parameter value itself is valid
                pass
            else:
                self.fail = True
                message = "Field {i} ({n}): {p} is not a valid parameter for function {f}".format(i=field_index,
                                                                                                  n=field_name,
                                                                                                  p=param_key,
                                                                                                  f=function_name)
                self.report.append(message)

    def check_for_required_params(self, field_name, field_index, function_name, valid_param, field_params):
        valid_parameter_key = valid_param["parameter"]
        if valid_parameter_key in field_params:
            pass
        else:
            self.fail = True
            message = "Field {i} ({n}): function {f} is missing required parameter {p}".format(i=field_index,
                                                                                               n=field_name,
                                                                                               f=function_name,
                                                                                               p=valid_parameter_key)
            self.report.append(message)


def load_validation(validation_type):
    validation_file = "src/resources/validation/{}-validation.yaml".format(validation_type)
    try:
        with open(validation_file, "r", encoding="utf-8") as f:
            validation = yaml.safe_load(f)

            return validation
    except FileNotFoundError:
        raise FileNotFoundError("No validation yaml found")


def print_report(report_type, report, fail, quiet, label=None):
    if not label:
        label = ""
    if fail:
        print("{} NOT VALID - REVIEW AND TRY AGAIN".format(report_type.upper()))
        report_filename = report_type + label + "validationreport.txt"
        with open(report_filename, "w+", encoding="utf-8") as f:
            f.write("\n".join(report))

    if not quiet:
        for line in report:
            print(line)



import yaml

from src.setup.Settings import read_config, write_config
from src.setup.Validator import MapValidator
from src.util.Output import Output


# Read the mapfile to generate outputs and populate them
# with fieldnames and rules for reharvests and processing
class Mapping():
    def __init__(self):
        self.mapping_file = read_config("mapping_file")
        self.outputs = []
        self.ordered_fieldnames = {}
        self.reharvest_triggers = []

        self.load_map()

    def load_map(self):
        # Read mapping from provided yaml document
        if not read_config("quiet"):
            print("Reading mapping file {}".format(self.mapping_file))

        map_data = None
        with open(self.mapping_file, "r", encoding="utf-8") as f:
            map_data = yaml.safe_load(f)

        for output_map in map_data:
            # Validate the map to ensure it contains required keys
            # And all fields have workable functions and parameters
            map_validator = MapValidator(output_map, read_config("quiet"), read_config("country_codes"))
            if map_validator.fail:
                exit("Map validation failed")

            # If all's well, create the output object for each map
            self.create_output_from_map(output_map)

    def create_output_from_map(self, output_map):
        # Set the label of the first output as the corefile
        label = output_map.get("label")
        if not read_config("corefile"):
            write_config("corefile", label)

        # Read output variables from map
        reference_column = output_map.get("reference_column")
        endpoint = output_map.get("primary_endpoint")
        explode_on = output_map.get("explode")
        reduce_on = output_map.get("reduce")
        requires = output_map.get("requires")
        output_rules = output_map.get("fields")

        fieldnames = []
        rules = []

        # For each field, read and save the rules for processing it
        for rule in output_rules:
            output_fieldname = list(rule.keys())[0]
            functions = rule.get(output_fieldname)
            new_rule = MappingRule(endpoint=endpoint, fieldname=output_fieldname, functions=functions)

            fieldnames.append(output_fieldname)
            rules.append(new_rule)

        # Create the Output object for this part of the map
        output = Output(label=label,
                        endpoint=endpoint,
                        reference_column=reference_column,
                        explode_on=explode_on,
                        reduce_on=reduce_on,
                        requires=requires,
                        fieldnames=fieldnames,
                        rules=rules)

        self.outputs.append(output)

        if output_map.get("extend"):
            self.create_triggers(output_map, endpoint)

# Use extension rules from the map to set up triggers that
# will fire when an applicable record is harvested
    def create_triggers(self, output_map, endpoint):
        for trigger in output_map.get("extend"):
            harvest_path = trigger.get("path")
            harvest_endpoint = trigger.get("endpoint")
            labels = trigger.get("for_label")
            if labels:
                labels = labels.split(", ")
            else:
                labels = [None]
            for label in labels:
                new_trigger = ReharvestTrigger(parent_endpoint=endpoint,
                                               harvest_path=harvest_path,
                                               harvest_endpoint=harvest_endpoint,
                                               label=label)

                self.reharvest_triggers.append(new_trigger)


class MappingRule():
    def __init__(self, endpoint=None, fieldname=None, functions=[]):
        self.endpoint = endpoint
        self.output_fieldname = fieldname
        self.mapping_functions = functions

        if not self.endpoint:
            self.endpoint = read_config("endpoint")

        # To add: validation of functions, do they have the parameters they should have

        if not read_config("quiet"):
            print("New mapping rule created for {e}, {o}".format(e=self.endpoint, o=self.output_fieldname))


class ReharvestTrigger():
    def __init__(self, parent_endpoint=None, harvest_path=None, harvest_endpoint=None, label=None):
        self.parent_endpoint = parent_endpoint
        self.harvest_path = harvest_path
        self.harvest_endpoint = harvest_endpoint
        self.label = label

        if not read_config("quiet"):
            print("New reharvest trigger created for {e}, {p}".format(e=self.harvest_endpoint, p=self.harvest_path))
